#!/usr/bin/perl -w 
$|=1;
use strict;

use IO::Socket::INET;
use Fcntl qw(F_GETFL F_SETFL O_NONBLOCK);
use Net::WebSocket::Server;
use JSON::Parse 'parse_json';
use Data::Dumper;

my $ws_port = 8090;
my $base_port = 10000;

print "Starting WebSocket chat server on port $ws_port, press Ctr-C to disconnect\n";

my @sockets;
my %connections;

sub add_listen($) {
	my $listen_port = $base_port + $_[0];
	print ("Listen: $listen_port\n");

	$sockets[$listen_port] = 1;

	$sockets[$listen_port] = IO::Socket::INET->new(LocalPort => $listen_port,
                                Type      => SOCK_STREAM,
                                Reuse     => 1,
                                Listen    => 100 )   # or SOMAXCONN
    or die "Couldn't be a tcp server on port $listen_port : $@\n";

	my $flags = fcntl($sockets[$listen_port] , F_GETFL, 0)
            or die "Can't get flags for the socket: $!\n";

	$flags = fcntl($sockets [$listen_port], F_SETFL, $flags | O_NONBLOCK)
            or die "Can't set flags for the socket: $!\n";

	return $sockets[$listen_port];
}


sub on_read {
	my $fd = $_[1]->accept();

	my $flags = fcntl($fd , F_GETFL, 0) or die "Can't get flags for the socket: $!\n";
	$flags = fcntl($fd, F_SETFL, $flags | O_NONBLOCK) or die "Can't set flags for the socket: $!\n";

	printf "Connect peer %s %s:%s \n", $fd->sockport(), $fd->peerhost(), $fd->peerport();
	if ($fd->connected()) {
		$connections{$fd->sockport()}{$fd->peerhost()}{conn}{$fd->peerport()} = $fd;
		if ($connections{$fd->sockport()}{$fd->peerhost()}{nmea}) {
			$fd->send($connections{$fd->sockport()}{$fd->peerhost()}{nmea});
		}
		$connections{$fd->sockport()}{$fd->peerhost()}{nmea} //= "\r\n";
	}
}

sub nmea_sum {
  my ($string) = @_;
  my $v = 0;
  $v ^= $_ for unpack 'C*', $string;
  sprintf '%02X', $v;
}

sub nmea_latlon($$){
	my ($lat,$lon) = @_;
	my $nmea = "";
	
	my $lata = abs($lat);
	my $latd = int($lata);
	my $latm = ($lata - $latd) * 60;
	my $lath = $lat > 0?"N":"S";

	my $lona = abs($lon);
	my $lond = int($lona);
	my $lonm = ($lona - $lond) * 60;
	my $lonh = $lon > 0?"E":"W";

    sprintf ("%02d%07.4f,%s,%03d%07.4f,%s", $latd,$latm,$lath,$lond,$lonm,$lonh);
}

sub nmea_time($) {
    my $time = $_[0];
    sprintf("%02d%02d%02d", (gmtime($time/1000))[2,1,0]);
}

sub nmea_date($) {
    my $time = $_[0];
    my @t = gmtime($time/1000);
    sprintf("%02d%02d%02d",$t[3],$t[4]+1, $t[5] %100);
}

sub format_nmea($$$$$$$) {
	my ($lat, $lon, $t, $speed, $hdg, $twa, $tws) = @_;

	$twa = 360 + $twa if $twa < 0;	

	my $mesg = sprintf('GNRMC,%s,A,%s,%0.2f,%0.1f,%s,,,A',
		nmea_time($t),
		nmea_latlon($lat, $lon),
		$speed,
		$hdg,
		nmea_date($t)
	);
	my $wind = sprintf('INMWV,%06.2f,T,%05.2f,N,A', $twa, $tws);
	return '$' . $mesg . '*' . nmea_sum($mesg) . "\r\n" .
		'$' . $wind . '*' . nmea_sum($wind) . "\r\n";
}
		
sub send_nmea($$) {
	my($ip, $json) = @_;
	my $error = 0;

	my $port = $json->{_id}{race_id} + $base_port;

	my $nmea = format_nmea($json->{pos}{lat},
					$json->{pos}{lon},
					$json->{lastCalcDate},
					$json->{speed},
					$json->{heading},
					$json->{twa},
					$json->{tws});

		foreach my $key (keys %{$connections{$port}{$ip}{conn}}) {
		my $fd = $connections{$port}{$ip}{conn}{$key};
		if (!$fd->connected || !$fd->send($nmea))
	 	{
			close($fd);
			delete $connections{$port}{$ip}{conn}{$key};
			print "Deleting $port $ip $key\n";
			$error = 1;
		}
	}
	$connections{$port}{$ip}{nmea} = $nmea;
}

sub resend_nmea() {
    foreach my $port (keys %connections) {
		foreach my $ip (keys %{$connections{$port}}) {
		 	foreach my $key (keys %{$connections{$port}{$ip}{conn}}) {
				if ($connections{$port}{$ip}{conn}{$key}->connected()) {
					my $res = $connections{$port}{$ip}{conn}{$key}->send($connections{$port}{$ip}{nmea});
					if (!$res) {
							#$connections{$port}{$ip}{$key}->close();
							delete $connections{$port}{$ip}{conn}{$key};
							print "Pruning $port $ip $key\n";
					}
		
				}
			}
		}
	}
}


Net::WebSocket::Server->new(
    listen => $ws_port,
    tick_period => 2,
    on_tick => sub {
		resend_nmea();
	},

    on_connect => sub {
        my ($serv, $conn) = @_;
		printf ("WS Connect from: %s\n", $conn->ip());
        $conn->on(
            utf8 => sub {
                my ($conn, $msg) = @_;
				my $json = parse_json($msg);
				if (!defined $sockets[$json->{_id}{race_id} + $base_port]) {
					my $socket = add_listen($json->{_id}{race_id});
					$serv->watch_readable(\*$socket => \&on_read);	
				} 
				send_nmea($conn->ip(), $json); #prepare the data feed

            },
        );
    },
)->start;
